from .readmol2 import Mol2, Mol2Block
from .readsmiles import Smiles
from .readpdb import PDB, PartialPDB, PDBBB
from .readmol import MolBlock
from .canonical_atom_order import CanonicalAtomOrderConverter



""" Covert molecule atoms to a canonical order based on canonical SMILES from RDKit
"""

from rdkit import Chem


class CanonicalAtomOrderConverter:
    def __init__(self, mol):
        """ Convert the molecule atom order to a canonical order.

        Args:
            mol (rdkit Mol object): The RDKit Mol object with any atom order.
        """
        self._mol = mol

    @property
    def mol(self):
        return self._mol

    def convert(self):
        """ Convert atom order.

        Returns:
            RDKit Mol object: An RDKit Mol object with canonical atom order.
        """
        # Creat canonical order dict
        old2new = Chem.CanonicalRankAtoms(
            self.mol, includeChirality=True, breakTies=True
        )
        new2old = {o: i for i, o in enumerate(old2new)}
        # build new molecule based on the new atom order
        new_mol = Chem.rdchem.RWMol(Chem.Mol())
        # add Atoms
        for idx in range(len(old2new)):
            new_mol.AddAtom(self.mol.GetAtomWithIdx(new2old[idx]))
        # rebuild Bonds
        bonds = self.mol.GetBonds()
        for b in bonds:
            new_mol.AddBond(
                old2new[b.GetBeginAtomIdx()],
                old2new[b.GetEndAtomIdx()],
                b.GetBondType(),
            )
        # Add conformer (atom 3D positions)
        try:
            old_conformer = self.mol.GetConformer(0)
        except ValueError:
            old_conformer = None
        if old_conformer is not None:
            new_conformer = Chem.Conformer(new_mol.GetNumAtoms())
            for idx in range(len(old2new)):
                pos = old_conformer.GetAtomPosition(new2old[idx])
                new_conformer.SetAtomPosition(idx, pos)
            new_mol.AddConformer(new_conformer)
        return new_mol



from rdkit.Chem.rdmolfiles import MolFromPDBFile, MolFromPDBBlock
import numpy as np
from scipy.spatial.distance import pdist
from scipy import sparse as sp
from scipy.linalg import toeplitz

from .basereader import _BaseReader
from ..utils.tools import property_getter


class PDB(_BaseReader):
    """ Read PDB file and convert to graph.
    """

    resi_to_int = {
        "ALA": 0,
        "CYS": 1,
        "ASP": 2,
        "GLU": 3,
        "PHE": 4,
        "GLY": 5,
        "HIS": 6,
        "ILE": 7,
        "LYS": 8,
        "LEU": 9,
        "MET": 10,
        "ASN": 11,
        "PRO": 12,
        "GLN": 13,
        "ARG": 14,
        "SER": 15,
        "THR": 16,
        "VAL": 17,
        "TRP": 18,
        "TYR": 19,
        "NAA": 20,  # not a amino acid
    }

    def __init__(self, fpath, sanitize=True):
        r"""
        smiles (str): smiles string
        """
        self._fpath = fpath
        self._sanitize = sanitize

    @classmethod
    def from_pdb_block(cls, block, sanitize=True):
        inst = cls(fpath=None, sanitize=sanitize)
        inst._rdkit_mol = MolFromPDBBlock(block)
        return inst

    @property
    @property_getter
    def num_atoms(self):
        """ Number of atoms
        """
        return self._num_atoms

    def _get_num_atoms(self):
        return self.rdkit_mol.GetNumAtoms()

    @property
    @property_getter
    def bonds(self):
        """ Bonds
        """
        return self._bonds

    def _get_bonds(self):
        bonds = list()
        for bond in self.rdkit_mol.GetBonds():
            b = dict()
            if bond.GetIsAromatic():
                type_ = "ar"
            else:
                type_ = str(int(bond.GetBondType()))
            b["connect"] = tuple([bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()])
            b["type"] = type_
            bonds.append(b)
        return bonds

    @property
    @property_getter
    def rdkit_mol(self):
        """ RDKit Mol object
        """
        return self._rdkit_mol

    def _get_rdkit_mol(self):
        self._rdkit_mol = MolFromPDBFile(self._fpath, sanitize=self._sanitize)
        return self._rdkit_mol

    @property
    @property_getter
    def atom_types(self):
        return self._atom_types

    def _get_atom_types(self):
        atom_types = list()
        for atom in self.rdkit_mol.GetAtoms():
            symbol = atom.GetSymbol().upper()
            atom_types.append(symbol)
        return atom_types

    def get_adjacency_matrix(self, sparse=False, padding=None):
        r""" Get the adjacency matrix of the molecular graph.

        Params:
            spase (bool): if True, return the matrix in sparse format
        =======================================================================
        Return:
            numpy.array or scipy.sparse.csc_matrix
        """
        if padding is None:
            matrix = np.zeros((self.num_atoms, self.num_atoms), dtype=np.int8)
        else:
            if padding < self.num_atoms:
                raise ValueError(
                    "Padding number should be larger than the atoms number."
                    "Got {} < {}".format(padding, self.num_atoms)
                )
            matrix = np.zeros((padding, padding), dtype=np.int8)
        for bond in self.bonds:
            edge = [c for c in bond["connect"]]
            matrix[edge, edge[::-1]] = 1
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def to_graph(self, sparse=False, pad_atom=None, pad_bond=None):
        graph = dict()
        graph["adjacency"] = self.get_adjacency_matrix(sparse=sparse, padding=pad_atom)
        graph["atom_features"] = self.get_atom_features(numeric=True, padding=pad_atom)
        graph["bond_features"] = self.get_bond_features(numeric=True)
        return graph

    def _is_atom(self, pdb_line):
        if pdb_line.startswith("ATOM"):
            return True
        if pdb_line.startswith("HETATM"):
            atom_type = pdb_line[17:20].strip().upper()
            if atom_type == "HOH":
                return False
            return True
        return False

    def get_atom_coordinates(self):
        """ Get atom coordinates from the raw pdb file.
        Return:
            list: list of atom coordinates with the same order as the pdb file.
        """
        atoms = list()
        with open(self._fpath, "r") as f:
            lines = f.readlines()
        for line in lines:
            if not self._is_atom(line):
                continue
            x = float(line[30:38])
            y = float(line[38:46])
            z = float(line[46:54])
            atoms.append((x, y, z))
        return atoms


class PartialPDB(PDB):
    def __init__(self, fpath, atom_list=None, cutoff=None, sanitize=True):
        super().__init__(fpath, sanitize=sanitize)
        self._atom_list = atom_list
        self._cutoff = cutoff

    @classmethod
    def from_pdb_block(cls, block, atom_list=None, cutoff=None, sanitize=True):
        inst = cls(fpath=None, atom_list=atom_list, cutoff=cutoff, sanitize=sanitize)
        inst._rdkit_mol = MolFromPDBBlock(block)
        return inst

    @property
    def atom_list(self):
        return self._atom_list

    @atom_list.setter
    def atom_list(self, value: list):
        assert isinstance(value, list)
        self._atom_list = value

    @property
    def cutoff(self):
        return self._cutoff

    @cutoff.setter
    def cutoff(self, value):
        assert isinstance(value, (int, float))
        self._cutoff = value

    @property
    @property_getter
    def coordinates(self):
        return self._coordinates

    def _get_coordinates(self):
        coor = np.zeros((self.num_atoms, 3))
        conf = self.rdkit_mol.GetConformer()
        for i in range(self.rdkit_mol.GetNumAtoms()):
            coor[i] = conf.GetAtomPosition(i)
        return coor

    def _pairwise_dist(self):
        try:
            n_atoms = len(self.atom_list)
        except TypeError:
            print("atom_list cannot be None. Initialize it with self.atom_list = [...]")
            raise
        dist_array = pdist(self.coordinates[self.atom_list])
        mat = np.zeros((n_atoms - 1, n_atoms - 1))
        i_upper = np.triu_indices(n_atoms - 1)
        mat[i_upper] = dist_array
        mat = np.pad(mat, [[0, 1], [1, 0]])
        i_lower = np.tril_indices(len(mat))
        mat[i_lower] = mat.T[i_lower]
        return mat

    def get_adjacency_matrix(self, sparse=False):
        r""" Get the adjacency matrix of the molecular graph with distance.
        spase (bool): if True, return the matrix in sparse format
        =======================================================================
        return (numpy.array or scipy.sparse.csc_matrix)
        """
        matrix = (self._pairwise_dist() <= self.cutoff).astype(int)
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def get_atom_features(self, numeric=False):
        r""" Get the atom features in the block. The feature contains
        coordinate and atom type for each atom.
        numeric (bool): if True, return the atom type as a number.
        =======================================================================
        return (list): list of tuples. The first three numbers in the tuples
            are coordinates and the last string or number is atom type.
        """
        features = list()
        atom_coors = list()
        atom_types = list()
        atom_degrees = list()
        atom_aromatic = list()
        atom_masses = list()
        for atom_id in self.atom_list:
            atom_coors.append(self.coordinates[atom_id])
            atom_types.append(self.atom_types[atom_id])
            atom = self.rdkit_mol.GetAtomWithIdx(atom_id)
            atom_degrees.append(atom.GetDegree())
            atom_aromatic.append(int(atom.GetIsAromatic()))
            atom_masses.append(atom.GetMass())
        for coor, typ, mass, deg, aro in zip(
            atom_coors, atom_types, atom_masses, atom_degrees, atom_aromatic
        ):
            if numeric:
                typ = self.atom_to_num(typ)
            features.append((*coor, typ, mass, deg, aro))
        return features

    def to_graph(self, sparse=False):
        graph = dict()
        graph["adjacency"] = self.get_adjacency_matrix(sparse=sparse)
        graph["atom_features"] = self.get_atom_features(numeric=True)
        return graph


class PDBBB(PartialPDB):
    """ Convert protein backbone to graph from PDB file.
    """

    _atom2int = {"C": 0, "N": 1}

    def __init__(self, fpath, sanitize=True):
        super().__init__(fpath=fpath, atom_list=None, cutoff=None, sanitize=sanitize)
        self._atom_list = self._get_backbone_atoms()

    def _get_backbone_atoms(self):
        atoms = list()
        for atom in self.rdkit_mol.GetAtoms():
            pdbinfo = atom.GetPDBResidueInfo()
            if pdbinfo.GetName().strip().upper() in ["N", "CA", "C"]:
                atoms.append(atom.GetIdx())
        return atoms

    def get_adjacency_matrix(self, sparse=False):
        r""" Get the adjacency matrix of the molecular graph with distance.
        spase (bool): if True, return the matrix in sparse format
        =======================================================================
        return (numpy.array or scipy.sparse.csc_matrix)
        """
        first_row = np.zeros((len(self.atom_list,)))
        first_row[[0, 1]] = 1
        first_col = np.zeros((len(self.atom_list,)))
        first_col[[0, 1]] = 1
        matrix = toeplitz(first_row, first_col)
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def get_atom_features(self, numeric=False):
        r""" Get the atom features in the block. The feature contains
        coordinate and atom type for each atom.
        Params:
            numeric (bool): if True, return the atom type as a number.
        =======================================================================
        Return:
            list of tuples. The first three numbers in the tuples
            are coordinates and the last string or number is atom type.
        """
        features = list()
        conf = self.rdkit_mol.GetConformer()
        for atom_id in self.atom_list:
            feat = list()
            feat += conf.GetAtomPosition(atom_id)
            atom = self.rdkit_mol.GetAtomWithIdx(atom_id)
            feat.append(self._atom2int[atom.GetSymbol()])
            pdbinfo = atom.GetPDBResidueInfo()
            try:
                feat.append(self.resi_to_int[pdbinfo.GetResidueName()])
            except KeyError:
                feat.append(self.resi_to_int["NAA"])
            features.append(feat)
        return features



import os
import shutil
from unittest import TestCase

from ..writers import GraphWriter
from ..readers import Mol2
from ..readers import Smiles
from ..readers import PDB, PartialPDB


class TestGraphWriting(TestCase):
    def setUp(self):
        self.outpath = os.path.join("tmp", "test_graph_writer")

    def tearDown(self):
        # remove created files
        shutil.rmtree(self.outpath)

    def test_generating_graphs_from_smiles(self):
        smiles = [
            "CC(=O)OC1=CC=CC=C1C(=O)O",  # Asprin
            "CCC(CC)COC(=O)[C@H](C)N[P@](=O)(OC[C@@H]1[C@H]([C@H]"
            "([C@](O1)(C#N)C2=CC=C3N2N=CN=C3N)O)O)"
            "OC4=CC=CC=C4",  # Remdesivir
        ]
        mols = [Smiles(s) for s in smiles]
        labels = [1, 0]

        writer = GraphWriter(mols)
        writer.write(self.outpath, prefix="test", graph_labels=labels)
        self.assertTrue(os.path.isfile(os.path.join(self.outpath, "test_A.txt")))
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_graph_indicator.txt"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_node_attributes.txt"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_edge_attributes.txt"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_graph_labels.txt"))
        )
        # assert adjacency matrices are correct
        with open(os.path.join(self.outpath, "test_A.txt"), "r") as f:
            lines = f.readlines()
        line1 = lines[0].split(",")
        adj1 = mols[0].get_adjacency_matrix(sparse=True).tocoo()
        self.assertEqual(int(line1[0]), adj1.row[0] + 1)
        self.assertEqual(int(line1[1]), adj1.col[0] + 1)
        n_bonds = mols[0].num_bonds
        n_atoms = mols[0].num_atoms
        line2 = lines[n_bonds * 2].split(",")  # Bonds are symmetric, so *2
        adj2 = mols[1].get_adjacency_matrix(sparse=True).tocoo()
        self.assertEqual(int(line2[0]), adj2.row[0] + n_atoms + 1)
        self.assertEqual(int(line2[1]), adj2.col[0] + n_atoms + 1)
        # assert graph indicator is correct
        with open(os.path.join(self.outpath, "test_graph_indicator.txt"), "r") as f:
            lines = f.readlines()
            self.assertEqual(lines.count("1\n"), mols[0].num_atoms)
            self.assertEqual(lines.count("2\n"), mols[1].num_atoms)
        # assert node labels are correct
        with open(os.path.join(self.outpath, "test_node_attributes.txt"), "r") as f:
            lines = f.readlines()
        true_feats = list()
        for mol in mols:
            true_feats += mol.get_atom_features(numeric=True)
        for line, tf in zip(lines, true_feats):
            for f1, f2 in zip(line.split(","), tf):
                self.assertEqual(float(f1), float(f2))
        # assert edge attributes are correct
        with open(os.path.join(self.outpath, "test_edge_attributes.txt"), "r") as f:
            attrs = f.readlines()
        with open(os.path.join(self.outpath, "test_A.txt"), "r") as f:
            adj = f.readlines()
        true_attrs = list()
        num_edges = list()
        num_nodes = list()
        for mol in mols:
            bond_featurs = mol.get_bond_features(numeric=True)
            true_attrs.append(bond_featurs)
            num_edges.append(len(bond_featurs))
            num_nodes.append(mol.num_atoms)
        t_nodes = 0
        t_edges = 0
        for tattr, ne, nn in zip(true_attrs, num_edges, num_nodes):
            counter = 0
            while counter < ne:
                at, ad = attrs[t_edges + counter], adj[t_edges + counter]
                key = "-".join(map(lambda x: str(int(x) - 1 - t_nodes), ad.split(",")))
                ea = tattr[key]
                self.assertEqual(ea, int(at))
                counter += 1
            t_nodes += nn
            t_edges += ne
        # assert graph labels are correct
        with open(os.path.join(self.outpath, "test_graph_labels.txt"), "r") as f:
            lines = f.readlines()
        for l1, l2 in zip(labels, lines):
            self.assertEqual(int(l1), int(l2))

    def test_generating_graphs_from_mol2blocks(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        test_gzip = os.path.join(dir_path, "testing_resources", "AAAARO.xaa.mol2.gz")
        mol = Mol2(test_gzip)
        mols = mol.mol2_blocks
        writer = GraphWriter(mols)
        writer.write(self.outpath, prefix="test", graph_labels=None)
        self.assertTrue(os.path.isfile(os.path.join(self.outpath, "test_A.txt")))
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_graph_indicator.txt"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_node_attributes.txt"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.outpath, "test_edge_attributes.txt"))
        )
        # assert adjacency matrices are correct
        with open(os.path.join(self.outpath, "test_A.txt"), "r") as f:
            lines = f.readlines()
        line1 = lines[0].split(",")
        adj1 = mols[0].get_adjacency_matrix(sparse=True).tocoo()
        self.assertEqual(int(line1[0]), adj1.row[0] + 1)
        self.assertEqual(int(line1[1]), adj1.col[0] + 1)
        n_bonds = mols[0].num_bonds
        n_atoms = mols[0].num_atoms
        line2 = lines[n_bonds * 2].split(",")  # Bonds are symmetric, so *2
        adj2 = mols[1].get_adjacency_matrix(sparse=True).tocoo()
        self.assertEqual(int(line2[0]), adj2.row[0] + n_atoms + 1)
        self.assertEqual(int(line2[1]), adj2.col[0] + n_atoms + 1)
        # assert graph indicator is correct
        with open(os.path.join(self.outpath, "test_graph_indicator.txt"), "r") as f:
            lines = f.readlines()
            self.assertEqual(lines.count("1\n"), mols[0].num_atoms)
            self.assertEqual(lines.count("2\n"), mols[1].num_atoms)
        # assert node labels are correct
        with open(os.path.join(self.outpath, "test_node_attributes.txt"), "r") as f:
            lines = f.readlines()
        true_feats = list()
        for mol in mols:
            true_feats += mol.get_atom_features(numeric=True)
        for line, tf in zip(lines, true_feats):
            for f1, f2 in zip(line.split(","), tf):
                self.assertEqual(float(f1), float(f2))
        # assert edge attributes are correct
        with open(os.path.join(self.outpath, "test_edge_attributes.txt"), "r") as f:
            attrs = f.readlines()
        with open(os.path.join(self.outpath, "test_A.txt"), "r") as f:
            adj = f.readlines()
        true_attrs = list()
        num_edges = list()
        num_nodes = list()
        for mol in mols:
            bond_featurs = mol.get_bond_features(numeric=True)
            true_attrs.append(bond_featurs)
            num_edges.append(len(bond_featurs))
            num_nodes.append(mol.num_atoms)
        t_nodes = 0
        t_edges = 0
        for tattr, ne, nn in zip(true_attrs, num_edges, num_nodes):
            counter = 0
            while counter < ne:
                at, ad = attrs[t_edges + counter], adj[t_edges + counter]
                key = "-".join(map(lambda x: str(int(x) - 1 - t_nodes), ad.split(",")))
                ea = tattr[key]
                self.assertEqual(ea, int(at))
                counter += 1
            t_nodes += nn
            t_edges += ne

    def test_prefix_bug_fix(self):
        smiles = [
            "CC(=O)OC1=CC=CC=C1C(=O)O",  # Asprin
            "CCC(CC)COC(=O)[C@H](C)N[P@](=O)(OC[C@@H]1[C@H]([C@H]"
            "([C@](O1)(C#N)C2=CC=C3N2N=CN=C3N)O)O)"
            "OC4=CC=CC=C4",  # Remdesivir
        ]
        mols = [Smiles(s) for s in smiles]
        labels = [1, 0]

        writer = GraphWriter(mols)
        writer.write(self.outpath, prefix=None, graph_labels=labels)
        self.assertTrue(os.path.isfile(os.path.join(self.outpath, "A.txt")))
        self.assertTrue(os.path.isfile(os.path.join(self.outpath, "graph_labels.txt")))

    def test_generating_graphs_from_pdb(self):
        fpath = os.path.join("chemreader", "tests", "testing_resources", "3CQW.pdb")
        pdb = PDB(fpath)
        part_pdb = PartialPDB(fpath, atom_list=list(range(10)), cutoff=1.5)
        writer = GraphWriter([pdb])
        writer.write(self.outpath, prefix="pdb", edge_features=False)
        writer = GraphWriter([part_pdb])
        writer.write(self.outpath, prefix="part_pdb", edge_features=False)



import os
import unittest
import logging

import numpy as np
from scipy import sparse as sp
from rdkit import Chem

from ..readers import Mol2, Mol2Block
from ..readers import Smiles
from ..readers import PDB, PartialPDB, PDBBB
from ..readers import CanonicalAtomOrderConverter
from ..readers.readmol import MolReader, MolBlock


class TestReadingMol2File(unittest.TestCase):
    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        test_gzip = os.path.join(dir_path, "testing_resources", "AAAARO.xaa.mol2.gz")
        self.mol = Mol2(test_gzip)
        self.blocks = self.mol.blocks
        self.first_block = self.blocks[0]
        test_f = os.path.join(dir_path, "testing_resources", "test_mol2_block")
        with open(test_f, "r") as f:
            self.block = Mol2Block(f.read())

    def test_read_blocks(self):
        self.assertTrue(len(self.blocks) > 0)
        self.assertTrue(isinstance(self.first_block, str))
        self.assertTrue(len(self.first_block) > 0)
        self.assertTrue("@<TRIPOS>" in self.first_block)

    def test_mol_name(self):
        self.assertEqual(self.block.mol_name, "ZINC000005319062")
        self.assertEqual(self.block._mol_name, "ZINC000005319062")

    def test_num_atoms(self):
        self.assertEqual(self.block.num_atoms, 28)
        self.assertEqual(self.block._num_atoms, 28)

    def test_num_bonds(self):
        self.assertEqual(self.block.num_bonds, 28)
        self.assertEqual(self.block._num_bonds, 28)

    def test_num_subst(self):
        self.assertEqual(self.block.num_subst, 1)
        self.assertEqual(self.block._num_subst, 1)

    def test_num_feat(self):
        self.assertEqual(self.block.num_feat, 0)
        self.assertEqual(self.block._num_feat, 0)

    def test_num_sets(self):
        self.assertEqual(self.block.num_sets, 0)
        self.assertEqual(self.block._num_sets, 0)

    def test_mol_type(self):
        self.assertEqual(self.block.mol_type, "SMALL")
        self.assertEqual(self.block._mol_type, "SMALL")

    def test_charge_type(self):
        self.assertEqual(self.block.charge_type, "USER_CHARGES")
        self.assertEqual(self.block._charge_type, "USER_CHARGES")

    def test_atom_names(self):
        self.assertEqual(len(self.block.atom_names), self.block.num_atoms)
        self.assertEqual(self.block.atom_names[0], "C")
        self.assertEqual(self.block.atom_names[-1], "H")
        self.assertTrue(hasattr(self.block, "_atom_names"))

    def test_coordinates(self):
        self.assertEqual(len(self.block.coordinates), self.block.num_atoms)
        self.assertTrue(
            np.array_equal(
                self.block.coordinates[0], np.array((-0.0178, 1.4648, 0.0101))
            )
        )
        self.assertTrue(
            np.array_equal(
                self.block.coordinates[-1], np.array((-1.3009, 0.3246, 7.4554))
            )
        )
        self.assertTrue(hasattr(self.block, "_coordinates"))

    def test_atom_types(self):
        self.assertEqual(len(self.block.atom_types), self.block.num_atoms)
        self.assertEqual(self.block.atom_types[0], "C")
        self.assertEqual(self.block.atom_types[-1], "H")
        self.assertEqual(self.block.atom_types[14], "N")
        self.assertTrue(hasattr(self.block, "_atom_types"))

    def test_atom_charges(self):
        self.assertEqual(len(self.block.atom_charges), self.block.num_atoms)
        self.assertEqual(self.block.atom_charges[0], -0.0600)
        self.assertEqual(self.block.atom_charges[-1], 0.4300)
        self.assertTrue(hasattr(self.block, "_atom_charges"))

    def test_bonds(self):
        self.assertEqual(len(self.block.bonds), self.block.num_bonds)
        self.assertTrue(hasattr(self.block, "_bonds"))
        bond = self.block.bonds[0]
        self.assertEqual(bond["connect"][0], 0)
        self.assertEqual(bond["connect"][1], 1)
        self.assertEqual(bond["type"], "1")
        bond = self.block.bonds[16]
        self.assertEqual(bond["connect"][0], 6)
        self.assertEqual(bond["connect"][1], 7)
        self.assertEqual(bond["type"], "am")


class TestBlockMissingInformation(unittest.TestCase):
    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        test_f = os.path.join(
            dir_path, "testing_resources", "test_mol2_block_missing_info"
        )
        with open(test_f, "r") as f:
            self.block = Mol2Block(f.read())

    def test_num_atoms(self):
        self.assertEqual(self.block.num_atoms, 28)
        self.assertEqual(self.block._num_atoms, 28)

    def test_num_subst(self):
        with self.assertLogs() as cm:
            self.block.num_subst
        self.assertIn("num_subst information is not available", cm.output[0])

    def test_num_feat(self):
        with self.assertLogs() as cm:
            self.block.num_feat
        self.assertIn("num_feat information is not available", cm.output[0])

    def test_num_sets(self):
        with self.assertLogs() as cm:
            self.block.num_sets
        self.assertIn("num_sets information is not available", cm.output[0])

    def test_atom_charges(self):
        with self.assertLogs() as cm:
            self.block.atom_charges
        self.assertIn("does not have charge information", cm.output[0])


class TestMol2(TestReadingMol2File):
    def test_mol2_to_smiles(self):
        can_smiles = self.mol.to_smiles()
        iso_smiles = self.mol.to_smiles(isomeric=True)
        self.assertEqual(len(can_smiles), self.mol.n_mols)
        self.assertEqual(can_smiles[0], r"C[NH+](C)CCNC(=O)c1nonc1N")
        self.assertEqual(len(iso_smiles), self.mol.n_mols)
        self.assertEqual(iso_smiles[62], r"Cn1cncc1[C@H]([NH3+])C1(O)CNC1")
        self.assertNotEqual(can_smiles[1], iso_smiles[1])

    def test_molecular_weights(self):
        mol_weights = self.mol.get_molecular_weights()
        self.assertEqual(len(mol_weights), self.mol.n_mols)
        for mw in mol_weights:
            self.assertGreater(mw, 0)
        self.assertEqual(int(mol_weights[0]), 200)

    def test_mol2blocks(self):
        mol2_blocks = self.mol.mol2_blocks
        self.assertEqual(len(mol2_blocks), self.mol.n_mols)
        self.assertTrue(isinstance(mol2_blocks[0], Mol2Block))

    def test_get_adjacency_matrices(self):
        matrices = self.mol.get_adjacency_matrices()
        self.assertEqual(len(matrices), self.mol.n_mols)
        self.assertTrue(isinstance(matrices[0], np.ndarray))
        self.assertEqual(np.sum(matrices[0]), self.mol.mol2_blocks[0].num_bonds * 2)
        self.assertEqual(matrices[0].shape, (28, 28))
        sparse_matrices = self.mol.get_adjacency_matrices(sparse=True)
        self.assertEqual(len(sparse_matrices), self.mol.n_mols)
        self.assertTrue(sp.issparse(sparse_matrices[0]))
        self.assertTrue(np.array_equal(sparse_matrices[0].toarray(), matrices[0]))

    def test_atom2int_and_bond2int(self):
        self.assertEqual(Mol2Block.atom_to_num("C"), 0)
        self.assertEqual(Mol2Block.atom_to_num("Any"), 22)
        self.assertEqual(Mol2Block.atom_to_num("@#$%"), 22)
        self.assertEqual(Mol2Block.bond_to_num("1"), 0)
        self.assertEqual(Mol2Block.bond_to_num("nc"), 6)
        self.assertEqual(Mol2Block.bond_to_num("@#$%"), 6)

    def test_get_atom_features(self):
        atom_features = self.mol.get_atom_features(numeric=False)
        self.assertEqual(len(atom_features), self.mol.n_mols)
        self.assertEqual(len(atom_features[0]), self.block.num_atoms)
        self.assertEqual(len(atom_features[0][0]), 7)
        self.assertTrue(isinstance(atom_features[0][0][0], float))
        self.assertTrue(isinstance(atom_features[0][0][3], str))
        numeric_features = self.mol.get_atom_features(numeric=True)
        self.assertEqual(len(numeric_features), self.mol.n_mols)
        self.assertEqual(len(numeric_features[0]), self.block.num_atoms)
        self.assertEqual(len(numeric_features[0][0]), 7)
        self.assertTrue(isinstance(numeric_features[0][0][0], float))
        self.assertTrue(isinstance(numeric_features[0][0][-1], int))

    def test_get_bond_features(self):
        bond_features = self.mol.get_bond_features(numeric=False)
        self.assertEqual(len(bond_features), self.mol.n_mols)
        self.assertEqual(len(bond_features[0]), self.block.num_bonds * 2)
        self.assertTrue(isinstance(bond_features[0]["1-2"], str))
        numeric_features = self.mol.get_bond_features(numeric=True)
        self.assertEqual(len(numeric_features), self.mol.n_mols)
        self.assertEqual(len(numeric_features[0]), self.block.num_bonds * 2)
        self.assertTrue(isinstance(numeric_features[0]["1-2"], int))

    def test_to_graphs(self):
        graphs = self.mol.to_graphs(sparse=False)
        self.assertEqual(len(graphs), self.mol.n_mols)
        self.assertTrue(isinstance(graphs[0]["adjacency"], np.ndarray))
        self.assertEqual(
            np.sum(graphs[0]["adjacency"]), self.mol.mol2_blocks[0].num_bonds * 2
        )
        self.assertEqual(graphs[0]["adjacency"].shape, (28, 28))
        self.assertEqual(len(graphs[0]["atom_features"]), self.block.num_atoms)
        self.assertEqual(len(graphs[0]["atom_features"][0]), 7)
        self.assertTrue(isinstance(graphs[0]["atom_features"][0][0], float))
        self.assertTrue(isinstance(graphs[0]["atom_features"][0][-1], int))
        self.assertEqual(len(graphs[0]["bond_features"]), self.block.num_bonds * 2)
        self.assertTrue(isinstance(graphs[0]["bond_features"]["1-2"], int))
        sparse_graphs = self.mol.to_graphs(sparse=True)
        self.assertEqual(len(sparse_graphs), self.mol.n_mols)
        self.assertTrue(sp.issparse(sparse_graphs[0]["adjacency"]))
        self.assertTrue(
            np.array_equal(
                sparse_graphs[0]["adjacency"].toarray(), graphs[0]["adjacency"]
            )
        )
        # graphs = self.mol.to_graphs(sparse=False, pad_atom=70, pad_bond=80)
        # self.assertEqual(graphs[0]["adjacency"].shape, (70, 70))
        # self.assertEqual(len(graphs[0]["atom_features"]), 70)
        # np.array(graphs[0]["atom_features"])
        # self.assertEqual(len(graphs[0]["bond_features"]), 80)
        # np.array(graphs[0]["bond_features"])
        # with self.assertRaises(ValueError):
        #     self.mol.to_graphs(sparse=False, pad_atom=27, pad_bond=80)
        # with self.assertRaises(ValueError):
        #     self.mol.to_graphs(sparse=False, pad_atom=70, pad_bond=27)


class TestReadingSmiles(unittest.TestCase):
    def setUp(self):
        # Aspirin
        self.smiles = "CC(=O)OC1=CC=CC=C1C(=O)O"
        self.sm = Smiles(self.smiles)

    def test_building_mol(self):
        self.assertTrue(hasattr(self.sm, "smiles_str"))
        self.assertEqual(self.sm.num_atoms, 13)
        self.assertEqual(self.sm.num_bonds, 13)

    def test_atom_types(self):
        types = self.sm.atom_types
        self.assertEqual(len(types), 13)
        self.assertEqual(types[0], "C")
        self.assertEqual(types[3], "O")
        self.assertEqual(types[1], "C")

    def test_bonds(self):
        bonds = self.sm.bonds
        self.assertEqual(len(bonds), 13)
        self.assertEqual(bonds[0]["type"], "1")
        self.assertEqual(bonds[0]["connect"], (0, 1))
        self.assertEqual(bonds[4]["type"], "ar")
        self.assertEqual(bonds[4]["connect"], (4, 5))

    def test_atom_featurs(self):
        feats = self.sm.get_atom_features()
        self.assertEqual(len(feats), 13)
        self.assertEqual(feats[0], ("C", 1, 0, 4, 0, 0))
        self.assertEqual(feats[4], ("C", 3, 0, 3, 1, 0))
        feats = self.sm.get_atom_features(numeric=True)
        self.assertEqual(len(feats), 13)
        self.assertEqual(feats[0], (0, 1, 0, 4, 0, 0))
        feats = self.sm.get_atom_features(numeric=True, padding=20)
        self.assertEqual(len(feats), 20)
        self.assertEqual(feats[0], (0, 1, 0, 4, 0, 0))
        self.assertEqual(feats[-1], (22, 0, 0, 0, 0, 0))
        with self.assertRaises(ValueError):
            self.sm.get_atom_features(padding=12)

    # def _sorted_correctly(self, atoms):
    #     """ To confirm the atoms are sorted correctly: same atoms are grouped together
    #     and sorted by their x coordinates.
    #     """

    def test_sorted_atoms(self):
        unsorted_atoms = self.sm.rdkit_mol.GetAtoms()
        # >>> [C, C, O, O, C, C, C, C, C, C, C, O, O]
        sorted_atoms = self.sm.sorted_atoms
        correct_sorted = ["C"] * 9 + ["O"] * 4
        for at, coat in zip(sorted_atoms, correct_sorted):
            self.assertEqual(at.GetSymbol(), coat)
        unsorted_indices = [at.GetIdx() for at in unsorted_atoms]
        sorted_indices = [at.GetIdx() for at in sorted_atoms]
        logging.debug("unsorted:" + str(unsorted_indices))
        logging.debug("sorted:" + str(sorted_indices))
        logging.debug("\n")
        unsorted_adj = self.sm.get_adjacency_matrix(sort_atoms=False)
        sorted_adj = self.sm.get_adjacency_matrix(sort_atoms=True)
        logging.debug("unsorted_adj:\n" + str(unsorted_adj))
        logging.debug("\n")
        logging.debug("sorted_adj:\n" + str(sorted_adj))

    def test_bond_features(self):
        feats = self.sm.get_bond_features()
        self.assertEqual(len(feats), 26)
        self.assertEqual(feats["0-1"], "1")
        feats = self.sm.get_bond_features(numeric=True)
        self.assertEqual(len(feats), 26)
        self.assertEqual(feats["0-1"], 0)
        # feats = self.sm.get_bond_features(numeric=True, padding=15)
        # self.assertEqual(len(feats), 15)
        # self.assertEqual(feats[0], 0)
        # self.assertEqual(feats[-1], 6)
        # with self.assertRaises(ValueError):
        #     self.sm.get_bond_features(padding=12)

    def test_sorted_bond_features(self):
        feats = self.sm.get_bond_features(sort_atoms=True)
        self.assertEqual(len(feats), 26)
        self.assertEqual(feats["0-5"], "1")
        with self.assertRaises(KeyError):
            feats["0-1"]
        unsorted_feats = self.sm.get_bond_features(sort_atoms=False)
        self.assertEqual(len(unsorted_feats), 26)
        self.assertEqual(unsorted_feats["0-1"], "1")

    def test_graph(self):
        graph = self.sm.to_graph()
        self.assertEqual(len(graph), 3)
        self.assertEqual(graph["adjacency"].shape, (13, 13))
        self.assertIsInstance(graph["adjacency"], np.ndarray)
        self.assertEqual(len(graph["atom_features"]), 13)
        self.assertEqual(len(graph["bond_features"]), 26)
        # graph = self.sm.to_graph(sparse=True, pad_atom=20, pad_bond=15)
        # self.assertIsInstance(graph["adjacency"], sp.csr_matrix)
        # self.assertEqual(graph["adjacency"].shape, (20, 20))
        # self.assertEqual(len(graph["atom_features"]), 20)
        # self.assertEqual(len(graph["bond_features"]), 15)

    def test_fingerprints(self):
        fp = self.sm.fingerprint
        self.assertEqual(len(fp), 2048)
        other = Smiles("C1ccccC1")
        self.assertFalse(self.sm.similar_to(other))
        self.assertTrue(self.sm.similar_to(self.sm))
        fp = Smiles("C1").fingerprint
        self.assertIsNone(fp)


class TestReadPDB(unittest.TestCase):
    def setUp(self):
        self.fpath = os.path.join(
            "chemreader", "tests", "testing_resources", "3CQW.pdb"
        )

    def test_mol_from_pdb_file(self):
        pdb = PDB(self.fpath)
        graph = pdb.to_graph()
        self.assertIn("adjacency", graph)
        self.assertIn("atom_features", graph)
        self.assertIn("bond_features", graph)

    def test_partial_pdb(self):
        al = [0, 1, 2, 3, 10]
        part_pdb = PartialPDB(self.fpath, atom_list=al)
        dist_mat = part_pdb._pairwise_dist()
        self.assertEqual(dist_mat.shape, (len(al), len(al)))
        part_pdb.cutoff = 1.5
        adj = part_pdb.get_adjacency_matrix()
        self.assertEqual(adj.shape, (len(al), len(al)))
        self.assertEqual(adj[0, 0], 1)
        self.assertEqual(adj[2, 1], 0)
        graph = part_pdb.to_graph()
        self.assertIn("adjacency", graph)
        self.assertIn("atom_features", graph)
        self.assertEqual(len(graph["atom_features"]), len(al))
        self.assertNotIn("bond_features", graph)
        conformer = part_pdb.rdkit_mol.GetConformer()
        self.assertEqual(
            graph["atom_features"][-1][:3], tuple(conformer.GetAtomPosition(10))
        )

    def test_coordinates(self):
        pdb = PDB(self.fpath, sanitize=False)
        counter = 0
        for atom in pdb.rdkit_mol.GetAtoms():
            counter += 1
        atoms = pdb.get_atom_coordinates()
        self.assertIsInstance(atoms, list)
        # self.assertEqual(len(atoms), pdb.num_atoms)

    def test_backbone_pdb(self):
        pdb = PDBBB(self.fpath, sanitize=False)
        adj = pdb.get_adjacency_matrix()
        self.assertTrue(np.array_equal(adj.diagonal(), np.ones(len(pdb.atom_list),)))
        self.assertTrue(
            np.array_equal(adj.diagonal(offset=1), np.ones(len(pdb.atom_list) - 1,))
        )
        self.assertTrue(
            np.array_equal(adj.diagonal(offset=-1), np.ones(len(pdb.atom_list) - 1,))
        )
        self.assertTrue(
            np.array_equal(adj.diagonal(offset=2), np.zeros(len(pdb.atom_list) - 2,))
        )
        self.assertTrue(
            np.array_equal(adj.diagonal(offset=-2), np.zeros(len(pdb.atom_list) - 2,))
        )
        atom_features = pdb.get_atom_features()
        # assert only backbone atoms are included in atom_features
        self.assertEqual(len(atom_features), 319 * 3)
        self.assertEqual(atom_features[-1][:3], [-14.909, -4.100, 8.772])
        self.assertEqual(len(atom_features), adj.shape[0])
        self.assertEqual(len(atom_features[0]), 5)


class TestReadMol(unittest.TestCase):
    def setUp(self):
        self.fpath = os.path.join(
            "chemreader", "tests", "testing_resources", "test_mol_reader.mol"
        )

    def test_block_reading(self):
        reader = MolReader(self.fpath)
        self.assertEqual(reader.n_mols, 3)
        self.assertIsInstance(reader.blocks, list)
        self.assertEqual(len(reader.blocks), 3)

        block = MolBlock(reader.blocks[0])
        self.assertIsInstance(block.rdkit_mol, Chem.rdchem.Mol)
        self.assertEqual(block.rdkit_mol.GetNumAtoms(), 24)
        self.assertEqual(block.rdkit_mol.GetNumBonds(), 26)
        adj = block.get_adjacency_matrix(sparse=False)
        self.assertEqual(adj.shape, (24, 24))
        sparse_adj = block.get_adjacency_matrix(sparse=True)
        self.assertIsInstance(sparse_adj, sp.csr.csr_matrix)
        atom_features = block.get_atom_features(numeric=False)
        self.assertEqual(len(atom_features), 24)
        self.assertEqual(atom_features[21][3], "Cl")
        atom_features = block.get_atom_features(numeric=True)
        self.assertEqual(atom_features[21][3], 7)


class TestAtomOrderConverter(unittest.TestCase):
    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.mol1 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol_origin.mol2"),
            removeHs=False,
        )
        self.mol2 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol_diff1.mol2"),
            removeHs=False,
        )
        self.mol3 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol_diff2.mol2"),
            removeHs=False,
        )
        self.mol4 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol2_origin.mol2"),
            removeHs=False,
        )
        self.mol5 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol2_diff1.mol2"),
            removeHs=False,
        )
        self.mol6 = Chem.MolFromMol2File(
            os.path.join(dir_path, "testing_resources", "mol2_diff2.mol2"),
            removeHs=False,
        )

    def assert_mols_equal(self, mol1, mol2):
        conf1 = mol1.GetConformer(0)
        conf2 = mol2.GetConformer(0)
        for idx in range(mol1.GetNumAtoms()):
            a1, a2 = (mol1.GetAtomWithIdx(idx), mol2.GetAtomWithIdx(idx))
            self.assertEqual(a1.GetSymbol(), a2.GetSymbol())
            self.assertEqual(a1.GetIdx(), a2.GetIdx())
            pos1 = conf1.GetAtomPosition(a1.GetIdx())
            pos2 = conf2.GetAtomPosition(a2.GetIdx())
            for coor1, coor2 in zip(pos1, pos2):
                self.assertEqual(coor1, coor2)

    def test_output_atom_order_are_the_same(self):
        conv1 = CanonicalAtomOrderConverter(self.mol1)
        conv2 = CanonicalAtomOrderConverter(self.mol2)
        conv3 = CanonicalAtomOrderConverter(self.mol3)

        new_mol1 = conv1.convert()
        new_mol2 = conv2.convert()
        new_mol3 = conv3.convert()

        self.assert_mols_equal(new_mol1, new_mol2)
        self.assert_mols_equal(new_mol1, new_mol3)



from rdkit import Chem

from .basereader import GraphFromRDKitMol


class Smiles(GraphFromRDKitMol):
    def __init__(self, smiles, sanitize=True):
        r"""
        smiles (str): smiles string
        """
        self._smiles_str = smiles
        self.sanitize = sanitize
        mol = Chem.MolFromSmiles(smiles)
        super().__init__(mol)

    @property
    def smiles_str(self):
        return self._smiles_str



def property_getter(func):
    """ Wrapper of the method that tries to return a hidden property of a
    instance. If the property if not defined, call the property constructor
    to generate the property and return it. The property name, hidden property
    name and constructor name must follow the rule below:
    property: foo
    hidden property: _foo
    constructor: _get_foo

    Example:
    class Foo:

        @property
        @property_getter
        def bar(self):
            return self._bar

        def _get_bar(self):
            code to construct the property
            return property_value
    """

    def wrapper(instance):
        try:
            return func(instance)
        except AttributeError:
            getter = instance.__getattribute__("_get_" + func.__name__)
            instance.__dict__["_" + func.__name__] = getter()
            return func(instance)

    return wrapper



import os


class _BaseWriter:
    def __init__(self):
        """
        """


class GraphWriter(_BaseWriter):
    def __init__(self, mols: list, **kwargs):
        super().__init__(**kwargs)
        self.mols = mols

    def write(self, out_path, prefix=None, edge_features=True, graph_labels=None):
        """ Write graphs to path.

        Args:
            out_path (str): Path to write out the graphs.
            prefix (str): Optional. Prefix of the output file names. Default is
                None.

        Returns:
            prefix_A.txt: adjacency matrices
            prefix_graph_indicator.txt: graph indicator
            prefix_node_label.txt: node features
            prefix_edge_attributes.txt: edge features
            prefix_graph_labels.txt: graph labels if graph_labels is not None
        """
        os.makedirs(out_path, exist_ok=True)
        prefix = "" if prefix is None else prefix + "_"
        a = open(os.path.join(out_path, prefix + "A.txt"), "w")
        idc = open(os.path.join(out_path, prefix + "graph_indicator.txt"), "w")
        n_label = open(os.path.join(out_path, prefix + "node_attributes.txt"), "w")
        if edge_features:
            edge_attr = open(
                os.path.join(out_path, prefix + "edge_attributes.txt"), "w"
            )
        # initialize variables for graph indicator and nodes indices
        graph_id = 1
        node_starting_index = 0
        for mol in self.mols:
            graph = mol.to_graph(sparse=True)
            adj = graph["adjacency"].tocoo()
            atom_feat = graph["atom_features"]
            if edge_features:
                bond_feat = graph["bond_features"]
            # write adjacency matrix and update node starting index
            for origin, target in zip(adj.row, adj.col):
                if edge_features:
                    # write bond features
                    bond = str(origin) + "-" + str(target)
                    writable_bond_feat = str(bond_feat[bond]) + "\n"
                    edge_attr.write(writable_bond_feat)
                origin = origin + 1 + node_starting_index
                target = target + 1 + node_starting_index
                a.write(str(origin) + "," + str(target) + "\n")
            node_starting_index += mol.num_atoms
            # write graph indicator
            idc.write((str(graph_id) + "\n") * mol.num_atoms)
            graph_id += 1
            # write node features
            writable_atom_feat = ""
            for feat in atom_feat:
                writable_atom_feat += ",".join(map(str, feat)) + "\n"
            n_label.write(writable_atom_feat)

        # close files
        a.close()
        idc.close()
        n_label.close()
        if edge_features:
            edge_attr.close()
        # write graph labels
        if graph_labels is not None:
            g_label = open(os.path.join(out_path, prefix + "graph_labels.txt"), "w")
            g_label.write("\n".join(map(str, graph_labels)))
            g_label.write("\n")
            g_label.close()



import gzip
import logging

from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt
import numpy as np
from scipy import sparse as sp
from tqdm import tqdm

from ..utils.tools import property_getter
from .basereader import _BaseReader


class Mol2Reader:
    """ Read .mol2 file and extract molecules in the file as blocks.
    """

    def __init__(self, path):
        """
        Args:
            path (str): path to the file
        """
        if path.endswith(".mol2"):
            with open(path, "r") as fh:
                self.file_contents = fh.readlines()
        elif path.endswith(".gz"):
            with gzip.open(path, "r") as fh:
                try:
                    lines = fh.readlines()
                    self.file_contents = [line.decode() for line in lines]
                except OSError:
                    logging.error("{} is not readble by gzip".format(path))
                    self.file_contents = None
        else:
            self.file_contents = None

    @property
    def n_mols(self):
        """
        Returns:
            int: Number of molecules in the file
        """
        if self.file_contents is None:
            return 0
        try:
            return self._n_mols
        except AttributeError:
            self._n_mols = 0
            for line in self.file_contents:
                if "@<TRIPOS>MOLECULE" in line:
                    self._n_mols += 1
            return self._n_mols

    @property
    @property_getter
    def blocks(self):
        """
        Returns:
            list: list of block contents as strings.
        """
        return self._blocks

    def _get_blocks(self):
        r""" Read the blocks in .mol2 file based on @<TRIPOS>MOLECULE label.
        return (list): list of block contents as strings.
        """
        if self.file_contents is None:
            return []
        block_starts = [
            i for i, l in enumerate(self.file_contents) if "@<TRIPOS>MOLECULE" in l
        ]
        if len(block_starts) == 0:
            return []
        blocks = list()
        i = 0
        while i + 1 < len(block_starts):
            block = "".join(self.file_contents[block_starts[i] : block_starts[i + 1]])
            blocks.append(block)
            i += 1
        blocks.append("".join(self.file_contents[block_starts[-1] :]))
        return blocks


class Mol2Block(_BaseReader):
    def __init__(self, block):
        r"""
        block (str): a mol2 format string of molecule block starting with
            @<TRIPOS>MOLECULE
        """
        self.block_str = block

    @property
    @property_getter
    def rdkit_mol(self):
        return self._rdkit_mol

    def _get_rdkit_mol(self):
        return Chem.MolFromMol2Block(self.block_str, sanitize=False)

    def _parse(self, block):
        r""" Parse the block content and dump the records into a dict
        """
        contents = block.strip().split("\n")
        contents_dict = dict()
        current_key = "<TEMP>"
        contents_dict[current_key] = list()
        for line in contents:
            if line.startswith("@"):
                try:
                    # get record type
                    current_key = line.split("<TRIPOS>")[1]
                    contents_dict[current_key] = list()
                    continue
                except IndexError:  # <TRIPOS> without a name
                    current_key = "<TEMP>"
                    continue
            contents_dict[current_key].append(line.strip())
        # discard the contents without record types
        contents_dict.pop("<TEMP>")
        return contents_dict

    @property
    @property_getter
    def block(self):
        return self._block

    def _get_block(self):
        return self._parse(self.block_str)

    @property
    @property_getter
    def mol_name(self):
        r""" Name of the molecule
        """
        return self._mol_name

    def _get_mol_name(self):
        return self.block["MOLECULE"][0]

    @property
    @property_getter
    def num_atoms(self):
        r""" Number of atoms in the molecule
        """
        return self._num_atoms

    def _get_num_atoms(self):
        return self.rdkit_mol.GetNumAtoms()

    @property
    @property_getter
    def num_bonds(self):
        return self._num_bonds

    def _get_num_bonds(self):
        return self.rdkit_mol.GetNumBonds()

    @property
    @property_getter
    def num_subst(self):
        return self._num_subst

    def _get_num_subst(self):
        try:
            return int(self.block["MOLECULE"][1].split()[2])
        except IndexError:
            logging.warning(
                "num_subst information is not " "available for {}".format(self.mol_name)
            )

    @property
    @property_getter
    def num_feat(self):
        return self._num_feat

    def _get_num_feat(self):
        try:
            return int(self.block["MOLECULE"][1].split()[3])
        except IndexError:
            logging.warning(
                "num_feat information is not " "available for {}".format(self.mol_name)
            )

    @property
    @property_getter
    def num_sets(self):
        return self._num_sets

    def _get_num_sets(self):
        try:
            return int(self.block["MOLECULE"][1].split()[4])
        except IndexError:
            logging.warning(
                "num_sets information is not " "available for {}".format(self.mol_name)
            )

    @property
    @property_getter
    def mol_type(self):
        return self._mol_type

    def _get_mol_type(self):
        return self.block["MOLECULE"][2]

    @property
    @property_getter
    def charge_type(self):
        return self._charge_type

    def _get_charge_type(self):
        return self.block["MOLECULE"][3]

    @property
    @property_getter
    def atom_names(self):
        return self._atom_names

    def _get_atom_names(self):
        atoms = self.rdkit_mol.GetAtoms()
        return [atom.GetSymbol() for atom in atoms]

    @property
    @property_getter
    def coordinates(self):
        return self._coordinates

    def _get_coordinates(self):
        conformer = self.rdkit_mol.GetConformer()
        return conformer.GetPositions()

    @property
    @property_getter
    def atom_types(self):
        return self._atom_types

    def _get_atom_types(self):
        atom_types = list()
        for atom in self.block["ATOM"]:
            type_ = atom.split()[5].split(".")[0].upper()
            atom_types.append(type_)
        return atom_types

    @property
    @property_getter
    def atom_charges(self):
        return self._atom_charges

    def _get_atom_charges(self):
        charges = list()
        for atom in self.block["ATOM"]:
            try:
                charge = atom.split()[8]
                charges.append(float(charge))
            except IndexError:
                logging.warning(
                    "{} does not have charge " "information.".format(self.mol_name)
                )
        return charges

    @property
    @property_getter
    def bonds(self):
        return self._bonds

    def _get_bonds(self):
        bonds = list()
        for bond in self.block["BOND"]:
            b = dict()
            tokens = bond.split()
            start = int(tokens[1]) - 1
            end = int(tokens[2]) - 1
            type_ = tokens[3]
            b["connect"] = tuple([start, end])
            b["type"] = type_
            bonds.append(b)
        return bonds

    @property
    @property_getter
    def molecular_weight(self):
        return self._molecular_weight

    def _get_molecular_weight(self):
        return ExactMolWt(self.rdkit_mol)

    def get_adjacency_matrix(self, sparse=False, padding=None):
        r""" Get the adjacency matrix of the molecular graph.
        spase (bool): if True, return the matrix in sparse format
        =======================================================================
        return (numpy.array or scipy.sparse.csc_matrix)
        """
        if padding is None:
            matrix = np.zeros((self.num_atoms, self.num_atoms), dtype=np.int8)
        else:
            if padding < self.num_atoms:
                raise ValueError(
                    "Padding number should be larger than the atoms number."
                    "Got {} < {}".format(padding, self.num_atoms)
                )
            matrix = np.zeros((padding, padding), dtype=np.int8)
        for bond in self.bonds:
            edge = [c for c in bond["connect"]]
            matrix[edge, edge[::-1]] = 1
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def get_atom_features(self, numeric=False, padding=None):
        r""" Get the atom features in the block. The feature contains
        coordinate and atom type for each atom.
        numeric (bool): if True, return the atom type as a number.
        =======================================================================
        return (list): list of tuples. The first three numbers in the tuples
            are coordinates and the last string or number is atom type.
        """
        features = list()
        atom_degrees = list()
        atom_aromatic = list()
        atom_masses = list()
        for atom in self.rdkit_mol.GetAtoms():
            atom_degrees.append(atom.GetDegree())
            atom_aromatic.append(int(atom.GetIsAromatic()))
            atom_masses.append(atom.GetMass())
        for coor, typ, mass, deg, aro in zip(
            self.coordinates, self.atom_types, atom_masses, atom_degrees, atom_aromatic
        ):
            if numeric:
                typ = self.atom_to_num(typ)
            features.append((*coor, typ, mass, deg, aro))
        if padding is not None:
            if padding < len(features):
                raise ValueError(
                    "Padding number should be larger than the feature number."
                    "Got {} < {}".format(padding, len(features))
                )
            pad = [(0.0, 0.0, 0.0, self.atom_to_num("ANY"), 0.0, 0, 0)] * (
                padding - len(features)
            )
            features.extend(pad)
        return features

    def to_smiles(self, isomeric=False):
        mol = Chem.RemoveHs(self.rdkit_mol)
        return Chem.MolToSmiles(mol, isomericSmiles=isomeric)

    def to_graph(self, sparse=False):
        graph = dict()
        graph["adjacency"] = self.get_adjacency_matrix(sparse=sparse)
        graph["atom_features"] = self.get_atom_features(numeric=True)
        graph["bond_features"] = self.get_bond_features(numeric=True)
        return graph


class Mol2(Mol2Reader):
    def __init__(self, path):
        super().__init__(path)

    @property
    def mol2_blocks(self):
        try:
            return self._mol2_blocks
        except AttributeError:
            self._mol2_blocks = self._get_mol2_blocks()
            return self._mol2_blocks

    def _get_mol2_blocks(self):
        m2blocks = list()
        for block in self.blocks:
            m2blocks.append(Mol2Block(block))
        return m2blocks

    def to_smiles(self, isomeric=False, verbose=0):
        r""" Convert the molecules in the file to SMILES strings
        isomeric (bool): False for cannonical, True for isomeric SMILES.
            Default is False.
        verbose (bool): Set to True to show progress bar.
        return (list): list of SMILES strings. If the molecule is not valid,
            an empty string will be added to the corresponding position in the
            list.
        """
        smiles = list()
        it = tqdm(self.mol2_blocks) if verbose else self.mol2_blocks
        for block in it:
            smiles.append(block.to_smiles(isomeric=isomeric))
        return smiles

    def get_molecular_weights(self):
        r""" Calculate the molecular weights
        return (list): list of molecular weights with the same order in the
            input file
        """
        mw = list()
        for block in self.mol2_blocks:
            mw.append(block.molecular_weight)
        return mw

    def get_adjacency_matrices(self, sparse=False):
        r""" Get adjacency matrices of the molecules as graphs
        sparse (bool): if to use sparse format for the matrices
        =======================================================================
        return (list): list of adjacency matrices with the same order as in the
            input file. The representations of the matrices are numpy arrays or
            numpy sparse matrices if the sparse argument is True.
        """
        matrices = list()
        for block in self.mol2_blocks:
            matrices.append(block.get_adjacency_matrix(sparse=sparse))
        return matrices

    def get_atom_features(self, numeric=False):
        r""" Get atom features (coordinates, atom type)
        numeric (bool): if True, return the atom types as numbers. The
            atoms that are able to be converted to consistant numbers are:
            C, N, O, S, F, Si, P, Cl, Br, Mg, Na, Ca, Fe, Al, I, B, K, Se, Zn,
            H, Cu, Mn. All other atom types will be treated as unknown and
            given a numeric type as 22.
        =======================================================================
        return (list): list of atom features in the same order as the input
            file.
        """
        atom_features = list()
        for block in self.mol2_blocks:
            atom_features.append(block.get_atom_features(numeric=numeric))
        return atom_features

    def get_bond_features(self, numeric=False):
        r""" Get bond features/types
        numeric (bool): if True, returen the bond types as numbers. Convertable
            bond types are "1,2,3,am,ar,du,un,nc". All other bond types will be
            treated as ANY and given a numeric type as 8.
        =======================================================================
        return (list): list of bond features in the same order as the input
            file.
        """
        bond_features = list()
        for block in self.mol2_blocks:
            bond_features.append(block.get_bond_features(numeric=numeric))
        return bond_features

    def to_graphs(self, sparse=False):
        r""" Convert the molecules to graphs that represented by atom features,
        bond types, and adjacency matrices.
        sparse (bool): if to use sparse format for the adjacency matrix
        """
        graphs = list()
        for block in self.mol2_blocks:
            graphs.append(block.to_graph(sparse=sparse))
        return graphs



import gzip
import logging

from rdkit import Chem
import numpy as np
from scipy import sparse as sp

from .basereader import _BaseReader
from ..utils.tools import property_getter


class MolReader:
    """ Read .mol file and extract molecules in the file as blocks.
    """

    def __init__(self, path):
        """
        Args:
            path (str): path to the file
        """
        if path.endswith(".mol"):
            with open(path, "r") as fh:
                self.file_contents = fh.readlines()
        elif path.endswith(".gz"):
            with gzip.open(path, "r") as fh:
                try:
                    lines = fh.readlines()
                    self.file_contents = [line.decode() for line in lines]
                except OSError:
                    logging.error("{} is not readble by gzip".format(path))
                    self.file_contents = None
        else:
            self.file_contents = None

    @property
    def n_mols(self):
        """
        Returns:
            int: Number of molecules in the file
        """
        if self.file_contents is None:
            return 0
        try:
            return self._n_mols
        except AttributeError:
            self._n_mols = 0
            for line in self.file_contents:
                if "M  END" in line:
                    self._n_mols += 1
            return self._n_mols

    @property
    @property_getter
    def blocks(self):
        """
        Returns:
            list: list of block contents as strings.
        """
        return self._blocks

    def _get_blocks(self):
        r""" Read the blocks in .mol2 file based on 'M  END' label.
        return (list): list of block contents as strings.
        """
        if self.file_contents is None:
            return []
        blocks = list()
        block = ""
        for line in self.file_contents:
            if "M  END" in line:
                blocks.append(block + line)
                block = ""
            else:
                block += line
        return blocks


class MolBlock(_BaseReader):
    r""" MDL Molfile block object.

    Args:
        block (str): a mol format string of molecule block ends with
            "M  END"
    """

    def __init__(self, block):
        self.block_str = block

    @property
    def num_atoms(self):
        return self.rdkit_mol.GetNumAtoms()

    @property
    def bonds(self):
        return self.rdkit_mol.GetBonds()

    @property
    def atom_types(self):
        return [atom.GetSymbol() for atom in self.rdkit_mol.GetAtoms()]

    @property
    @property_getter
    def rdkit_mol(self):
        return self._rdkit_mol

    def _get_rdkit_mol(self):
        return Chem.MolFromMolBlock(self.block_str, sanitize=False)

    def get_adjacency_matrix(self, sparse=False, padding=None):
        r""" Get the adjacency matrix of the molecular graph.
        spase (bool): if True, return the matrix in sparse format
        =======================================================================
        return (numpy.array or scipy.sparse.csc_matrix)
        """
        num_atoms = self.rdkit_mol.GetNumAtoms()
        if padding is None:
            matrix = np.zeros((num_atoms, num_atoms), dtype=np.int8)
        else:
            if padding < num_atoms:
                raise ValueError(
                    "Padding number should be larger than the atoms number."
                    "Got {} < {}".format(padding, num_atoms)
                )
            matrix = np.zeros((padding, padding), dtype=np.int8)
        for bond in self.rdkit_mol.GetBonds():
            begin, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            matrix[(begin, end)] = 1
            matrix[(end, begin)] = 1
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def get_atom_features(self, numeric=False):
        r""" Get the atom features in the block. The feature contains
        coordinate and atom type for each atom.
        numeric (bool): if True, return the atom type as a number.
        =======================================================================
        return (list): list of tuples. The first three numbers in the tuples
            are coordinates and the last string or number is atom type.
        """
        features = list()
        conformer = self.rdkit_mol.GetConformer(0)
        for atom in self.rdkit_mol.GetAtoms():
            coors = conformer.GetAtomPosition(atom.GetIdx())
            atom_type = atom.GetSymbol()
            if numeric:
                atom_type = self.atom_to_num(atom_type)
            features.append([coors.x, coors.y, coors.z, atom_type])
        return features

    def to_smiles(self, isomeric=False):
        mol = Chem.RemoveHs(self.rdkit_mol)
        return Chem.MolToSmiles(mol, isomericSmiles=isomeric)

    def to_graph(self, sparse=False):
        graph = dict()
        graph["adjacency"] = self.get_adjacency_matrix(sparse=sparse)
        graph["atom_features"] = self.get_atom_features(numeric=True)
        return graph



from copy import deepcopy

from abc import ABCMeta, abstractmethod, abstractproperty
import numpy as np
from rdkit.Chem.Descriptors import ExactMolWt
from rdkit.Chem import AllChem
from rdkit import DataStructs
from scipy import sparse as sp

from ..utils.tools import property_getter


class _BaseReader(metaclass=ABCMeta):

    # https://github.com/shionhonda/gae-dgl/blob/master/gae_dgl/prepare_data.py
    _avail_atom_types = [
        "C",
        "N",
        "O",
        "S",
        "F",
        "Si",
        "P",
        "Cl",
        "Br",
        "Mg",
        "Na",
        "Ca",
        "Fe",
        "Al",
        "I",
        "B",
        "K",
        "Se",
        "Zn",
        "H",
        "Cu",
        "Mn",
        "unknown",
    ]

    _atom2int = {atom.upper(): idx for idx, atom in enumerate(_avail_atom_types)}

    _bond_types = ["1", "2", "3", "am", "ar", "du", "un"]
    _bond2int = {bond.upper(): idx for idx, bond in enumerate(_bond_types)}

    @classmethod
    def atom_to_num(cls, atom_type):
        return cls._atom2int.get(atom_type.upper(), cls._atom2int["UNKNOWN"])

    @classmethod
    def bond_to_num(cls, bond_type):
        return cls._bond2int.get(bond_type.upper(), cls._bond2int["UN"])

    @staticmethod
    def rebuild_adj(adj, new_idx):
        """ Rebuld adjacency matrix with the new indices.
        Args:
            adj (numpy 2D array or matrix): The adjacency matrix to rebuild.
            new_idx (list of int): The list of new indices of the old nodes. For
                example, an old adjacency matrix with 3 nodes changes its first node
                index to 1 and second node index to 0. The new_idx should be [1, 0, 2].
        Returns:
            numpy 2D array or matrix: The rebuilt adjacency matrix.
        """
        new_idx = {old: new for new, old in enumerate(new_idx)}
        new_adj = np.zeros(adj.shape, dtype=np.int)
        for row in range(adj.shape[0]):
            for col in range(adj.shape[1]):
                if adj[row, col] == 0:
                    continue
                new_r = new_idx[row]
                new_c = new_idx[col]
                new_adj[new_r, new_c] = 1
        return new_adj

    def one_of_k_encoding(self, x, allowable_set):
        if x not in allowable_set:
            raise Exception(
                "input {0} not in allowable set{1}:".format(x, allowable_set)
            )
        return list(map(lambda s: x == s, allowable_set))

    def one_of_k_encoding_unk(self, x, allowable_set):
        """Maps inputs not in the allowable set to the last element."""
        if x not in allowable_set:
            x = allowable_set[-1]
        return list(map(lambda s: x == s, allowable_set))

    @abstractproperty
    def num_atoms(self):
        """ Number of atoms
        """

    @abstractproperty
    def bonds(self):
        """ Bonds
        """

    @abstractproperty
    def rdkit_mol(self):
        """ RDKit Mol object
        """

    @abstractproperty
    def atom_types(self):
        """ Atom types
        """

    @abstractmethod
    def get_adjacency_matrix(self):
        """ Get the adjacency matrix
        """

    @property
    def sorted_atoms(self):
        try:
            return self._sorted_atoms
        except AttributeError:
            self._sortAtoms(self.rdkit_mol.GetAtoms())
            return self._sorted_atoms

    def _sortAtoms(self, atoms):
        def key(atom):
            type_ = self._atom2int.get(atom.GetSymbol(), len(self._atom2int))
            degree = atom.GetDegree()
            idx = atom.GetIdx()
            return (type_, degree, idx)

        self._sorted_atoms = sorted(atoms, key=key)

    def get_atom_features(self, numeric=False, sort_atoms=False, padding=None):
        r""" Get the atom features in the block. The feature contains
        coordinate and atom type for each atom.
        Args:
            numeric (bool): if True, return the atom type as a number.
            sort_atoms (bool): Default is False. If True, sort the atoms by atom type.
            padding (None or int): Pad atom feature matrix to a fix length. The number
                must be larger than the number of atoms in the molecules.

        Returns:
            list: list of tuples. Features are atom type, atom mass, atom
                degree, and atom aromatic
        """
        features = list()
        if sort_atoms:
            atoms = self.sorted_atoms
        else:
            atoms = self.rdkit_mol.GetAtoms()
        for i, atom in enumerate(atoms):
            feature = list()
            # the features of an atom includes: atom type, degree, formal charge,
            # hybridization, aromatic, and chirality
            atom_type = self.atom_types[i]
            if numeric:
                atom_type = self.atom_to_num(atom_type)
            feature.append(atom_type)
            feature.append(atom.GetDegree())
            # feature.append(atom.GetImplicitValence())
            feature.append(atom.GetFormalCharge())
            # feature.append(atom.GetNumRadicalElectrons())
            feature.append(int(atom.GetHybridization()))
            feature.append(int(atom.GetIsAromatic()))
            feature.append(int(atom.GetChiralTag()))
            features.append(tuple(feature))
        if padding is not None:
            if padding < len(features):
                raise ValueError(
                    "Padding number should be larger than the feature number."
                    "Got {} < {}".format(padding, len(features))
                )
            pad = (
                [tuple([self.atom_to_num("unknown")] + [0] * (len(features[0]) - 1))]
            ) * (padding - len(features))
            features.extend(pad)
        return features

    def sort_bonds(self, unsorted_bonds):
        """ Sort bonds based on sorted atoms.
        Args:
            unsorted_bonds (list): list of bonds in chemical compound.

        Returns:
            dict: Bond feature dict.
        """
        new_idx = {old.GetIdx(): new for new, old in enumerate(self.sorted_atoms)}
        sorted_bonds = list()
        for bond in unsorted_bonds:
            start, end = bond["connect"]
            new_bond = deepcopy(bond)
            new_bond["connect"] = [0, 0]
            new_bond["connect"][0] = new_idx[start]
            new_bond["connect"][1] = new_idx[end]
            sorted_bonds.append(new_bond)
        return sorted_bonds

    def get_bond_features(self, numeric=False, sort_atoms=False):
        r""" Get the bond features/types in the block.
        numeric (bool): if True, return the bond type as a number.
        =======================================================================
        return (list): list of bond types.
        """
        features = dict()
        if sort_atoms:
            bonds = self.sort_bonds(self.bonds)
        else:
            bonds = self.bonds
        for bond in bonds:
            type_ = bond["type"]
            conn = str(bond["connect"][0]) + "-" + str(bond["connect"][1])
            conn2 = str(bond["connect"][1]) + "-" + str(bond["connect"][0])
            if numeric:
                type_ = self.bond_to_num(type_)
            features[conn] = type_
            features[conn2] = type_
        return features

    @abstractmethod
    def to_graph(self):
        """ Convert molecule to graph
        """


class GraphFromRDKitMol(_BaseReader):
    def __init__(self, mol):
        r"""
        Args:
            mol (rdkit Mol object)
        """
        self._rdkit_mol = mol

    @property
    def rdkit_mol(self):
        return self._rdkit_mol

    @property
    @property_getter
    def num_atoms(self):
        r""" Number of atoms in the molecule
        """
        return self._num_atoms

    def _get_num_atoms(self):
        return self.rdkit_mol.GetNumAtoms()

    @property
    @property_getter
    def num_bonds(self):
        return self._num_bonds

    def _get_num_bonds(self):
        return self.rdkit_mol.GetNumBonds()

    @property
    @property_getter
    def atom_names(self):
        return self._atom_names

    def _get_atom_names(self):
        atoms = self.rdkit_mol.GetAtoms()
        return [atom.GetSymbol() for atom in atoms]

    @property
    @property_getter
    def atom_types(self):
        return self._atom_types

    def _get_atom_types(self):
        atom_types = list()
        for atom in self.rdkit_mol.GetAtoms():
            symbol = atom.GetSymbol().upper()
            atom_types.append(symbol)
        return atom_types

    @property
    @property_getter
    def fingerprint(self):
        return self._fingerprint

    def _get_fingerprint(self):
        if self.rdkit_mol is None:
            return None
        fingerprint = AllChem.GetMorganFingerprintAsBitVect(self.rdkit_mol, 2)
        return fingerprint

    @property
    @property_getter
    def bonds(self):
        return self._bonds

    def _get_bonds(self):
        bonds = list()
        for bond in self.rdkit_mol.GetBonds():
            b = dict()
            if bond.GetIsAromatic():
                type_ = "ar"
            else:
                type_ = str(int(bond.GetBondType()))
            b["connect"] = tuple([bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()])
            b["type"] = type_
            bonds.append(b)
        return bonds

    @property
    @property_getter
    def molecular_weight(self):
        return self._molecular_weight

    def _get_molecular_weight(self):
        return ExactMolWt(self.rdkit_mol)

    def get_adjacency_matrix(self, sparse=False, sort_atoms=False, padding=None):
        r""" Get the adjacency matrix of the molecular graph.
        spase (bool): if True, return the matrix in sparse format
        =======================================================================
        return (numpy.array or scipy.sparse.csc_matrix)
        """
        if padding is None:
            matrix = np.zeros((self.num_atoms, self.num_atoms), dtype=np.int8)
        else:
            if padding < self.num_atoms:
                raise ValueError(
                    "Padding number should be larger than the atoms number."
                    "Got {} < {}".format(padding, self.num_atoms)
                )
            matrix = np.zeros((padding, padding), dtype=np.int8)
        for bond in self.bonds:
            edge = [c for c in bond["connect"]]
            matrix[edge, edge[::-1]] = 1
        if sort_atoms:
            matrix = self.rebuild_adj(matrix, [at.GetIdx() for at in self.sorted_atoms])
        if sparse:
            matrix = sp.csr_matrix(matrix)
        return matrix

    def to_graph(self, sparse=False, sort_atoms=False, pad_atom=None, pad_bond=None):
        graph = dict()
        graph["adjacency"] = self.get_adjacency_matrix(
            sparse=sparse, sort_atoms=sort_atoms, padding=pad_atom
        )
        graph["atom_features"] = self.get_atom_features(
            numeric=True, sort_atoms=sort_atoms, padding=pad_atom
        )
        graph["bond_features"] = self.get_bond_features(
            numeric=True, sort_atoms=sort_atoms
        )
        return graph

    def similar_to(self, other, threshold=0.5):
        sim = DataStructs.FingerprintSimilarity(self.fingerprint, other.fingerprint)
        if sim > threshold:
            return True
        return False



from setuptools import setup, find_packages

VERSION = "0.3.10"

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="chemreader",
    version=VERSION,
    author="Yang Liu",
    author_email="thomasliuy@gmail.com",
    description="Read data from typical chemical file formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thomasly/chem_reader",
    download_url=f"https://github.com/thomasly/chem_reader/archive/{VERSION}.tar.gz",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy", "scipy"],
    python_requires=">=3.6",
)
